#ifndef _NVFS_H
#define _NVFS_H

#include <linux/fs.h>
#include <linux/mount.h>
#include <linux/statfs.h>
#include <linux/falloc.h>
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/vmalloc.h>
#include <linux/dax.h>
#include <linux/pfn_t.h>
#include <linux/random.h>
#include <linux/uio.h>
#include <linux/iomap.h>
#include <linux/libnvdimm.h>
#include <linux/seq_file.h>
#include <linux/parser.h>
#include <linux/xattr.h>
#include <linux/posix_acl_xattr.h>
#include <linux/security.h>
#include <linux/quotaops.h>
#include <linux/compat.h>
#include <asm/unaligned.h>
#include <asm/word-at-a-time.h>

#include "struct.h"

#define NVFS_USR_QUOTA		1
#define NVFS_GRP_QUOTA		2
#define NVFS_DAX		4
#define NVFS_DAX_AUTO		8

struct nvfs_per_cpu {
	unsigned long current_inode;
	unsigned long current_block;
};

struct nvfs_group_info {
	unsigned n_free_blocks;
	unsigned n_free_inodes;
	spinlock_t lock;
} ____cacheline_aligned_in_smp;

struct nvfs_superblock {
	struct super_block *s;
	struct dax_device *dax_dev;
	uint64_t memory_map_size;
	void *memory_map;
	bool memory_vmapped;

	u8 log2_page_size;
	u8 log2_inode_size;
	u8 log2_bytes_per_group;
	u8 log2_inodes_per_group;

	unsigned mount_flags;

	uint64_t n_groups;

	struct nvfs_per_cpu __percpu *npc;

	struct nvfs_group_info *group_info;

	struct percpu_counter free_blocks_cnt;
	struct percpu_counter free_inodes_cnt;

	u8 saved_dirty;
};

struct nvfs_memory_inode {
	struct inode vfs_inode;
	struct nvfs_inode *nvi;
	struct rw_semaphore dax_sem;
	struct rw_semaphore xattr_sem;
	int was_mmapped;
	uint32_t i_flag;
	struct timespec64 i_btime;
#ifdef CONFIG_QUOTA
	struct dquot *i_dquot[MAXQUOTAS];
#endif
};

/*
 * Some older processors have inefficient cache flushing - for them, we don't
 * want to use cache flusing at all - so we should define CONFIG_USE_NT_STORES.
 *
 * For Skylake-X, CONFIG_USE_NT_STORES should not be defined because it degrades
 * performance.
 */
//#define CONFIG_USE_NT_STORES
#if 1
#define nv_store(dest, src)				\
do {							\
	typeof(dest) uniq = (src);			\
	memcpy_flushcache(&(dest), &uniq, sizeof(dest));\
} while (0)
#else
#define nv_store nv_store_cached
#endif

#ifdef CONFIG_USE_NT_STORES
#define nv_store_cached(dest, src)	nv_store(dest, src)
#define nv_flush_cached(ptr, len)	do { } while (0)
#else
#define nv_store_cached(dest, src)			\
do {							\
	(dest) = (src);					\
} while(0)
#define nv_flush_cached			arch_wb_cache_pmem
#endif


static inline unsigned long log2_pages_per_group(struct nvfs_superblock *nvs)
{
	return nvs->log2_page_size + 3;
}

static inline unsigned long log2_group_size(struct nvfs_superblock *nvs)
{
	return nvs->log2_page_size * 2 + 3;
}

static inline struct nvfs_memory_inode *i_to_nmi(struct inode *inode)
{
	return container_of(inode, struct nvfs_memory_inode, vfs_inode);
}

static inline struct nvfs_sb *p_superblock(struct nvfs_superblock *nvs)
{
	return (struct nvfs_sb *)nvs->memory_map;
}

static inline struct nvfs_group *p_group(struct nvfs_superblock *nvs, unsigned long group)
{
	return (struct nvfs_group *)((char *)nvs->memory_map + (group << log2_group_size(nvs)) + (2UL << nvs->log2_page_size));
}

static inline uint8_t *p_bitmap(struct nvfs_superblock *nvs, unsigned long group)
{
	return (char *)nvs->memory_map + (group << log2_group_size(nvs)) + (1UL << nvs->log2_page_size);
}

static inline struct nvfs_inode *p_inode(struct nvfs_superblock *nvs, ino_t ino)
{
	size_t offset = (ino & ((1UL << nvs->log2_inodes_per_group) - 1)) << nvs->log2_inode_size;
	offset += 2UL << nvs->log2_page_size;
	offset += ino >> nvs->log2_inodes_per_group << nvs->log2_bytes_per_group;
	return (struct nvfs_inode *)((char *)nvs->memory_map + offset);
}

static inline bool nvfs_inode_has_indirect_blocks(struct inode *inode)
{
	return	S_ISREG(inode->i_mode) ||
		(S_ISLNK(inode->i_mode) && !(i_to_nmi(inode)->i_flag & NVFS_FLAG_FAST_SYMLINK));
}

static inline void nvfs_inc_link_count(struct inode *inode)
{
	inc_nlink(inode);
	nv_store(i_to_nmi(inode)->nvi->i_links, cpu_to_le32(inode->i_nlink));
}

static inline void nvfs_dec_link_count(struct inode *inode)
{
	drop_nlink(inode);
	nv_store(i_to_nmi(inode)->nvi->i_links, cpu_to_le32(inode->i_nlink));
}

static inline unsigned nvfs_max_filename_length(struct nvfs_superblock *nvs)
{
	return (1UL << nvs->log2_page_size) - NVFS_DIR_OFFSET - offsetof(struct nvfs_dir_entry, name);
}

#define FS_REG_FLMASK (~(FS_DIRSYNC_FL | FS_TOPDIR_FL))
#define FS_OTHER_FLMASK (FS_NODUMP_FL | FS_NOATIME_FL)
static inline uint32_t nvfs_mask_flags(umode_t mode, uint32_t flags)
{
	if (S_ISDIR(mode))
		return flags;
	else if (S_ISREG(mode))
		return flags & FS_REG_FLMASK;
	else
		return flags & FS_OTHER_FLMASK;
}

/* acl.c */

#ifdef CONFIG_FS_POSIX_ACL
#define CONFIG_NVFS_POSIX_ACL
#endif

#ifdef CONFIG_NVFS_POSIX_ACL
struct posix_acl *nvfs_get_acl(struct inode *inode, int type);
int nvfs_set_acl(struct inode *inode, struct posix_acl *acl, int type);
int nvfs_init_acl(struct inode *inode, struct inode *dir);
#else
#define nvfs_get_acl	NULL
#define nvfs_set_acl	NULL
static inline int nvfs_init_acl(struct inode *inode, struct inode *dir) { return 0; }
#endif

/* balloc.c */

void nvfs_free_blocks(struct nvfs_superblock *nvs, unsigned long i, unsigned long n);
unsigned long nvfs_free_run(struct nvfs_superblock *nvs, __le64 *run, unsigned long n, bool zero);
uint64_t nvfs_new_block(struct nvfs_superblock *nvs, unsigned *prealloc);

/* bmap.c */

void nvfs_set_inode_size(struct nvfs_memory_inode *nmi, uint64_t size);
void nvfs_set_inode_blocks(struct nvfs_memory_inode *nmi, uint64_t blocks);
void nvfs_zero_page(struct nvfs_superblock *nvs, void *ptr, unsigned prealloc, void *aux);
void *nvfs_bmap(struct nvfs_memory_inode *nmi, uint64_t page, unsigned *prealloc,
		void (*create_page)(struct nvfs_superblock *nvs, void *page, unsigned prealloc, void *aux), void *aux, bool *new);
int nvfs_punch_pages(struct nvfs_memory_inode *nmi, uint64_t start_page, uint64_t end_page, bool zero);
int nvfs_punch_hole(struct nvfs_memory_inode *nmi, loff_t start, loff_t end);

/* dentry.c */

extern const struct dentry_operations nvfs_dops;

/* dir.c */

int nvfs_init_empty_dir(struct inode *dir);
struct nvfs_dir_entry *nvfs_prepare_add(struct dentry *dentry, struct inode *inode);
void nvfs_finish_add(struct nvfs_dir_entry *nde, struct dentry *dentry, struct inode *inode);
int nvfs_add_link(struct dentry *dentry, struct inode *inode);
struct nvfs_dir_entry *nvfs_find_de(struct nvfs_memory_inode *nmi, struct qstr *qname);
int nvfs_delete_de(struct nvfs_memory_inode *nmi, struct qstr *qname);
int nvfs_is_dir_empty(struct nvfs_memory_inode *nmi);
void nvfs_delete_directory(struct inode *dir);

extern const struct inode_operations nvfs_dir_inode_operations;
extern const struct file_operations nvfs_dir_fops;

/* file.c */

#ifdef CONFIG_QUOTA
ssize_t nvfs_quota_read(struct super_block *sb, int type, char *buf, size_t len, loff_t off);
ssize_t nvfs_quota_write(struct super_block *sb, int type, const char *data, size_t len, loff_t off);
#endif

ssize_t nvfs_rw_iter_locked(struct kiocb *iocb, struct iov_iter *iov);

extern const struct inode_operations nvfs_file_inode_operations;
extern const struct file_operations nvfs_file_fops;

/* filecache.c */

int nvfs_cache_readpage(struct file *file, struct page *page);
int nvfs_cache_writepage(struct page *page, struct writeback_control *wbc);
int nvfs_cache_write_begin(struct file *file, struct address_space *mapping, loff_t pos, unsigned len, unsigned flags, struct page **pagep, void **fsdata);
int nvfs_cache_write_end(struct file *file, struct address_space *mapping, loff_t pos, unsigned len, unsigned copied, struct page *pagep, void *fsdata);

ssize_t nvfs_cached_rw_locked(struct nvfs_memory_inode *nmi, struct file *file, bool wr, char __user *buf, size_t len, loff_t *ppos);

/* ialloc.c */

void nvfs_free_inode(struct nvfs_superblock *nvs, unsigned long i);
struct inode *nvfs_new_inode(struct inode *dir, umode_t mode, dev_t rdev, const struct qstr *qstr);

/* inode.c */

void nvfs_set_inode_flags(struct inode *inode);
void nvfs_fill_inode(struct inode *inode, struct nvfs_inode *nvi);
struct inode *nvfs_iget(struct super_block *sb, unsigned long ino);
void nvfs_evict_inode(struct inode *inode);
int nvfs_getattr(const struct path *path, struct kstat *stat, uint32_t request_mask, unsigned int query_flags);
int nvfs_setattr(struct dentry *dentry, struct iattr *attr);
int nvfs_update_time(struct inode *inode, struct timespec64 *time, int flags);

/* ioctl.c */

long nvfs_ioctl(struct file *filp, unsigned cmd, unsigned long arg);
#ifdef CONFIG_COMPAT
long nvfs_compat_ioctl(struct file *file, unsigned cmd, unsigned long arg);
#endif

/* journal.c */

int nvfs_init_journal(struct nvfs_superblock *nvs);

/* mmap.c */

int nvfs_mmap(struct file *file, struct vm_area_struct *vma);

/* namei.c */

int nvfs_create(struct inode *dir, struct dentry *dentry, umode_t mode, bool excl);
int nvfs_mkdir(struct inode *dir, struct dentry *dentry, umode_t mode);
int nvfs_symlink(struct inode *dir, struct dentry *dentry, const char *symname);
int nvfs_mknod(struct inode *dir, struct dentry *dentry, umode_t mode, dev_t rdev);
int nvfs_unlink(struct inode *dir, struct dentry *dentry);
int nvfs_rmdir(struct inode *dir, struct dentry *dentry);
int nvfs_rename(struct inode *old_dir, struct dentry *old_dentry, struct inode *new_dir, struct dentry *new_dentry, unsigned flags);
int nvfs_link(struct dentry *old_dentry, struct inode *dir, struct dentry *dentry);

/* super.c */

void nvfs_error(struct nvfs_superblock *nvs, const char *fmt, ...) __printf(2, 3);

/* utils.c */

void nvfs_memzero_flushcache(char *ptr, size_t len);
uint64_t nvfs_get_random(uint64_t limit);
unsigned long nvfs_find_bit_in_bmp(uint8_t *bitmap, unsigned long start, unsigned long end);

/* xattr.c */

int nvfs_xattr_get(struct inode *inode, int flags, const char *name,
		   void *buffer, size_t buffer_len);
int nvfs_xattr_set(struct inode *inode, int type, const char *name,
		   const void *buffer, size_t buffer_len, int flags);

ssize_t nvfs_listxattr(struct dentry *dentry, char *list, size_t list_size);

int nvfs_init_security(struct inode *inode, struct inode *dir,
		       const struct qstr *qstr);

extern const struct xattr_handler *nvfs_xattr_handlers[];

#endif
