#ifndef _NVFS_STRUCT_H
#define _NVFS_STRUCT_H

/*
Layout:
	superblock		1 page
	block bitmap		1 page
	group descriptor	the end is not aligned on page boundary, the rest is inode bitmap
	inode bitmap		the end is not aligned on page boundary, the rest are inodes
	inodes			n pages
	data blocks		m pages
*/

#define DEFAULT_BYTES_PER_INODE	20480

#define NVFS_SB_STATIC_PART	2048
#define NVFS_MAGIC		"NVFS0.13"
#define NVFS_MAGIC_LEN		8

#define NVFS_UUID_LEN		16

#define MIN_LOG_PAGE_SIZE	12
#define MAX_LOG_PAGE_SIZE	21	/* due to NVFS_XATTR_TOTAL_LEN_MASK */

#define MIN_LOG_INODE_SIZE	8

#define NVFS_DIRECT_BLOCKS	11
#define NVFS_INDIRECT_BLOCKS	5
#define NVFS_FAST_SYMLINK_LEN	((NVFS_DIRECT_BLOCKS + NVFS_INDIRECT_BLOCKS) * 8)
#define NVFS_RADIX_BLOCKS	((NVFS_DIRECT_BLOCKS + NVFS_INDIRECT_BLOCKS) / 2)
#define NVFS_LOG2_RADIX_BLOCKS	__builtin_ctz(NVFS_RADIX_BLOCKS)
#define NVFS_MAX_TREE_DEPTH	4

#define NVFS_JOUNAL_LENGTH	NVFS_RADIX_BLOCKS

struct nvfs_journal_entry {
	__le64 address;
	__le64 value;
};

#define NVFS_JOURNAL_VALID		0x8000000000000000ULL
#define NVFS_JOURNAL_TOP_VALUE_BIT	0x4000000000000000ULL
#define NVFS_JOURNAL_END		0xffffffffffffffffULL

struct nvfs_journal {
	struct nvfs_journal_entry entries[NVFS_JOUNAL_LENGTH + 1];
};

struct nvfs_journal_block {
	__le64 next;
	struct nvfs_journal j[];
};

struct nvfs_sb {
	union {
		struct {
			__u8 magic[NVFS_MAGIC_LEN];
			__u8 uuid[NVFS_UUID_LEN];
			__le64 n_pages;
			__le64 root_inode;
			__u8 log2_page_size;
			__u8 log2_inode_size;
			__u8 log2_inode_pages_per_group;
			__u8 checksum;
		};
		__u8 sb_static_part_align[NVFS_SB_STATIC_PART];
	};
	union {
		struct {
			__le32 dirty;
			__le32 error;
			__u8 sb_journal_align[56];
			struct nvfs_journal_block inline_journal;
		};
	};
};

struct nvfs_group {
	__le32 n_free_blocks;
	__le32 n_free_inodes;
	__u8 bitmap[];
};

#define NVFS_XATTR_PREFIX_SHIFT			13
#define NVFS_XATTR_PREFIX_EMPTY			0x0000
#define NVFS_XATTR_PREFIX_USER			0x2000
#define NVFS_XATTR_PREFIX_TRUSTED		0x6000
#define NVFS_XATTR_PREFIX_SECURITY		0xa000
#define NVFS_XATTR_PREFIX_POSIX_ACL_DEFAULT	ACL_TYPE_DEFAULT	/* 0x4000 */
#define NVFS_XATTR_PREFIX_POSIX_ACL_ACCESS	ACL_TYPE_ACCESS		/* 0x8000 */

#define NVFS_XATTR_TOTAL_LEN_MASK		0x001fffff
#define NVFS_XATTR_PREFIX_MASK			0x00e00000
#define NVFS_XATTR_NAMELEN_MASK			0xff000000

#define NVFS_XATTR_TOTAL_LEN_OFFSET		0
#define NVFS_XATTR_PREFIX_OFFSET		21
#define NVFS_XATTR_NAMELEN_OFFSET		24

struct nvfs_xattr_entry {
	__le32 x;
	__u8 data[];
};

static inline unsigned nvfs_xattr_get_total_len(struct nvfs_xattr_entry *e)
{
	return (e->x & NVFS_XATTR_TOTAL_LEN_MASK) >> NVFS_XATTR_TOTAL_LEN_OFFSET;
}

static inline __u8 nvfs_xattr_get_name_len(struct nvfs_xattr_entry *e)
{
	return (e->x & NVFS_XATTR_NAMELEN_MASK) >> NVFS_XATTR_NAMELEN_OFFSET;
}

static inline unsigned nvfs_xattr_get_prefix(struct nvfs_xattr_entry *e)
{
	return (e->x & NVFS_XATTR_PREFIX_MASK) >> (NVFS_XATTR_PREFIX_OFFSET - NVFS_XATTR_PREFIX_SHIFT);
}

static inline size_t nvfs_xattr_entry_needed_size(size_t name_len, size_t data_len)
{
	size_t s = name_len + data_len;
	s = (s + 3) & ~3;
	return sizeof(struct nvfs_xattr_entry) + s;
}

static inline size_t nvfs_xattr_entry_size(const struct nvfs_xattr_entry *nxe)
{
	size_t s = (le32_to_cpu(nxe->x) & NVFS_XATTR_TOTAL_LEN_MASK) >> NVFS_XATTR_TOTAL_LEN_OFFSET;
	s = (s + 3) & ~3;
	return sizeof(struct nvfs_xattr_entry) + s;
}

static inline __le32 nvfs_xattr_entry_make_size(size_t s)
{
	return cpu_to_le32((s - sizeof(struct nvfs_xattr_entry)) << NVFS_XATTR_TOTAL_LEN_OFFSET);
}

struct nvfs_inode {
	__le64 i_size;
	__le64 i_blocks;
	union {
		struct {
			__le64 i_direct_blocks[NVFS_DIRECT_BLOCKS];
			__le64 i_indirect_blocks[NVFS_INDIRECT_BLOCKS];
		};
		__le64 i_rdev;
		__u8 i_fast_symlink[NVFS_FAST_SYMLINK_LEN];
		__le64 i_radix[2][NVFS_RADIX_BLOCKS];
	};
	__le64 i_xattr_block;
	__le64 i_ctime;
	__le64 i_mtime;
	__le64 i_atime;
	__le64 i_btime;
	__le32 i_ctime_nsec;
	__le32 i_mtime_nsec;
	__le32 i_atime_nsec;
	__le32 i_btime_nsec;
	__le32 i_uid;
	__le32 i_gid;
	__le32 i_links;
	__le32 i_mode;
	__le32 i_flag;
	struct nvfs_xattr_entry i_xattr[];
};

#define NVFS_FLAGS			(FS_COMPR_FL | FS_SYNC_FL | FS_IMMUTABLE_FL | FS_APPEND_FL | FS_NODUMP_FL | FS_NOATIME_FL | FS_DIRSYNC_FL | FS_TOPDIR_FL)
#define NVFS_FLAGS_INHERITED		(FS_COMPR_FL | FS_SYNC_FL | FS_NODUMP_FL | FS_NOATIME_FL | FS_DIRSYNC_FL)
#define NVFS_FLAG_SECOND_DIR_BANK	0x40000
#define NVFS_FLAG_FAST_SYMLINK		0x400000

struct nvfs_dir_entry {
	__le64 inode;
	__le32 name_len_and_type;
	__u8 name[];
};

#define NAME_LEN_MASK	0x00ffffff
#define NAME_LEN_OFFSET	0
#define TYPE_MASK	0xff000000
#define TYPE_OFFSET	24

static inline size_t nvfs_dir_entry_namelen(const struct nvfs_dir_entry *nde)
{
	return (le32_to_cpu(nde->name_len_and_type) & NAME_LEN_MASK) >> NAME_LEN_OFFSET;
}

static inline unsigned char nvfs_dir_entry_type(const struct nvfs_dir_entry *nde)
{
	return (le32_to_cpu(nde->name_len_and_type) & TYPE_MASK) >> TYPE_OFFSET;
}

static inline size_t nvfs_dir_entry_needed_size(size_t name_len)
{
	return (name_len + 19) & ~7;
}

static inline size_t nvfs_dir_entry_size(const struct nvfs_dir_entry *nde)
{
	return nvfs_dir_entry_needed_size(nvfs_dir_entry_namelen(nde));
}

static inline size_t nvfs_dir_entry_make_size(size_t s)
{
	return s - 16;
}

struct nvfs_dir_head {
	__le64 magic;
	__le64 self;
	__le64 next;
	__le64 inode;
	struct nvfs_dir_entry nde[];
};

#define NVFS_DIR_HEAD_MAGIC		cpu_to_le64(0xff65646f6e726964ULL)
#define NVFS_DIR_OFFSET			offsetof(struct nvfs_dir_head, nde)

static inline unsigned nvfs_max_chain_length(unsigned log2_page_size)
{
	return (1U << (32 - (log2_page_size - 4))) - 1;
}


#define GR_64 0x61C8864680B583EBULL

static inline uint32_t nvfs_hash_name(const unsigned char *name, size_t len)
{
	uint64_t x = 0, y = 0;
	while ((ssize_t)len > 0) {
#ifdef CONFIG_DCACHE_WORD_ACCESS
		uint64_t a = le64_to_cpu(load_unaligned_zeropad(name));
		if (len < 8) {
			a &= (1ULL << (len * 8)) - 1;
		}
#else
		uint64_t a;
		if (len >= 8) {
			a = get_unaligned_le64(name);
		} else {
			size_t s;
			a = 0;
			for (s = 0; s < len; s++) {
				a |= (uint64_t)name[s] << (s * 8);
			}
		}
#endif

		x ^= a;
		y ^= x;
		x = rol64(x, 12);
		x += y;
		y = rol64(y, 45);
		y *= 9;

		name += 8;
		len -= 8;
	}

	y ^= x * GR_64;
	y *= GR_64;
	return y >> 32;
}


static inline const char *nvfs_validate_super(const struct nvfs_sb *sb, uint64_t size)
{
	uint8_t log2_pages_per_group;
	uint8_t checksum;
	unsigned i;

	if (size < 4096)
		return "the device is too small";
	if (memcmp(sb->magic, NVFS_MAGIC, NVFS_MAGIC_LEN))
		return "superblock not found";

	checksum = 0;
	for (i = 0; i < NVFS_SB_STATIC_PART; i++)
		checksum += ((uint8_t *)sb)[i];
	if (checksum)
		return "invalid checksum";

	if (sb->log2_page_size < MIN_LOG_PAGE_SIZE || sb->log2_page_size > MAX_LOG_PAGE_SIZE)
		return "invalid page size";
	if (sb->log2_inode_size < MIN_LOG_INODE_SIZE || sb->log2_inode_size > sb->log2_page_size)
		return "invalid inode size";

	log2_pages_per_group = sb->log2_page_size + 3;
	if (sb->log2_inode_pages_per_group >= log2_pages_per_group)
		return "invalid inode pages per group";

	if (size >> sb->log2_page_size < le64_to_cpu(sb->n_pages))
		return "the device is too small";

	return NULL;
}

static inline void nvfs_get_sizes(const struct nvfs_sb *sb,
				  uint64_t *p_n_groups,
				  uint64_t *p_n_inodes,
				  uint64_t *p_inodes_in_group,
				  uint64_t *p_inodes_in_last_group,
				  uint64_t *p_inode_bitmap_inodes,
				  uint64_t *p_max_size)
{
	uint8_t log2_pages_per_group;
	uint64_t n_groups, remaining_pages;
	uint64_t inodes_per_group, inode_bitmap_bytes, inode_bitmap_inodes;
	uint64_t inodes_in_group, inodes_in_last_group;

	log2_pages_per_group = sb->log2_page_size + 3;

	inodes_per_group = 1U << sb->log2_page_size >> sb->log2_inode_size << sb->log2_inode_pages_per_group;
	inode_bitmap_bytes = (inodes_per_group + 7) / 8;
	inode_bitmap_inodes = (sizeof(struct nvfs_group) + inode_bitmap_bytes + ((1U << sb->log2_inode_size) - 1)) >> sb->log2_inode_size;

	n_groups = le64_to_cpu(sb->n_pages) >> log2_pages_per_group;
	remaining_pages = le64_to_cpu(sb->n_pages) & ((1ULL << log2_pages_per_group) - 1);

	inodes_in_group = inodes_per_group - inode_bitmap_inodes;

	if (remaining_pages >= 3) {
		uint64_t i;
		n_groups++;
		inodes_in_last_group = 0;
		for (i = inode_bitmap_inodes; i < inodes_per_group; i++) {
			uint64_t offs = (2 << sb->log2_page_size) + (i << sb->log2_inode_size);
			offs >>= sb->log2_page_size;
			if (offs < remaining_pages)
				inodes_in_last_group++;
			else
				break;
		}
	} else {
		inodes_in_last_group = inodes_in_group;
	}

	if (p_n_groups)
		*p_n_groups = n_groups;
	if (p_n_inodes)
		*p_n_inodes = inodes_in_group * (n_groups - 1) + inodes_in_last_group;
	if (p_inodes_in_group)
		*p_inodes_in_group = inodes_in_group;
	if (p_inodes_in_last_group)
		*p_inodes_in_last_group = inodes_in_last_group;
	if (p_inode_bitmap_inodes)
		*p_inode_bitmap_inodes = inode_bitmap_inodes;

	if (p_max_size) {
		uint64_t max_pages;
		unsigned i;

		max_pages = NVFS_DIRECT_BLOCKS;
		for (i = 0; i < NVFS_INDIRECT_BLOCKS; i++) {
			uint64_t val = LLONG_MAX;
			unsigned shift = (sb->log2_page_size - 3) * (i + 1);
			if (shift < 64) {
				val = 1ULL << shift;
			}
			if (max_pages + val < val)
				max_pages = ULLONG_MAX;
			else
				max_pages += val;
		}
		if (max_pages << sb->log2_page_size >> sb->log2_page_size == max_pages)
			*p_max_size = max_pages << sb->log2_page_size;
		else
			*p_max_size = ULLONG_MAX;
		if (*p_max_size > (uint64_t)LLONG_MAX - ((1L << sb->log2_page_size) - 1))
			*p_max_size = (uint64_t)LLONG_MAX - ((1L << sb->log2_page_size) - 1);
	}
}

#endif
